{******************************************************************************}
{ Jazarsoft FormEx Component                                                   }
{******************************************************************************}
{                                                                              }
{ VERSION      : 1.0                                                           }
{ AUTHOR       : James Azarja                                                  }
{ CREATED      : 10 July 2000                                                  }
{ WEBSITE      : http://www.jazarsoft.cjb.net/                                 }
{ SUPPORT      : support@jazarsoft.cjb.net                                     }
{ BUG-REPORT   : bugreport@jazarsoft.cjb.net                                   }
{ COMMENT      : comment@jazarsoft.cjb.net                                     }
{ LEGAL        : Copyright (C) 2000 Jazarsoft.                                 }
{                                                                              }
{******************************************************************************}
{ NOTE         :                                                               }
{                                                                              }
{ This code may be used and modified by anyone so long as  this header and     }
{ copyright  information remains intact.                                       }
{                                                                              }
{ The code is provided "as-is" and without warranty of any kind,               }
{ expressed, implied or otherwise, including and without limitation, any       }
{ warranty of merchantability or fitness for a  particular purpose.           }
{                                                                              }
{ In no event shall the author be liable for any special, incidental,          }
{ indirect or consequential damages whatsoever (including, without             }
{ limitation, damages for loss of profits, business interruption, loss         }
{ of information, or any other loss), whether or not advised of the            }
{ possibility of damage, and on any theory of liability, arising out of        }
{ or in connection with the use or inability to use this software.           }
{                                                                              }
{******************************************************************************}
	
unit FormEx;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ShellApi, Registry, Menus, Extctrls;  

Const
  SysMenuExID          = $FFF;
  ScrollCaptionTimerID = 12345;
  TrayIconTimerID      = 54321;
  WM_ICONTRAYNOTIFY    = WM_USER + 1234;
  IconID               = 12345;


type
  TCycleEvent      = procedure(Sender: TObject; Current: Integer) of object;

  TDrawMethod      = (dmNormal, dmCenter, dmTile, dmStretch);
  TDropFilesEvent  = procedure(Files: TStrings; const Pos: TPoint) of object;
  TOnMovingEvent   = procedure (Sender: TObject;var Rect:TRect) of object;

  TScrollDirection = (dLeft, dRight);

  THoundred = 0..100;

  TTransparency = class(TPersistent)
  private
    FTranspColor     : TColor;
    FTransparency    : THoundred;
    ScreenShoot      : TBitmap;
    ParentForm       : TForm;
    Design           : Boolean;
    FEnabled         : Boolean;
    CanTakeScreen    : Boolean;

    procedure TakeScreenShoot;

    procedure SetTranspColor(Value: TColor);
    procedure SetTransparency(Value: THoundred);
  protected
  public
    constructor Create(PForm : tForm);
    destructor Destroy; override;
  published
    property TranspColor      : TColor    read FTranspColor  write SetTranspColor;
    property Transparency     : THoundred read FTransparency write SetTransparency;
    property Enabled          : Boolean   read FEnabled      write FEnabled;
  end;

  TTrayIcon = class(TPersistent)
  private
    ParentForm     : tForm;
    ParentComp     : TComponent;
    Design         : Boolean;

    FEnabled       : Boolean;
    FIcon          : TIcon;
    FIconVisible   : Boolean;
    FHint          : String;
    FShowHint      : Boolean;
    FLeftPopupMenu : TPopupMenu;
    FRightPopupMenu: TPopupMenu;

    FIconList      : TImageList;
    FCycleIcons    : Boolean;
    FCycleInterval : Cardinal;
    IconIndex      : Integer;           
    procedure SetCycleIcons(Value: Boolean);
    procedure SetCycleInterval(Value: Cardinal);
    procedure HandleIconMessage(var Msg: TMessage);
    function  InitIcon: Boolean;
    procedure SetIcon(Value: TIcon);
    procedure SetIconVisible(Value: Boolean);
    procedure SetHint(Value: String);
    procedure SetShowHint(Value: Boolean);
    procedure PopupAtCursor(Index:Integer);
  protected
    IconData                  : TNotifyIconData;
    procedure Click; dynamic;
    procedure DblClick; dynamic;
    procedure CycleIcon; dynamic;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); dynamic;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); dynamic;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); dynamic;

    function ShowIcon: Boolean; virtual;
    function HideIcon: Boolean; virtual;
    function ModifyIcon: Boolean; virtual;
  public
    constructor Create(PForm:tForm;PComp:tComponent;InDesign:Boolean);
    destructor Destroy; override;
    Procedure NextIconCycle;
  published
    property IconList             : TImageList  read FIconList        write FIconList;
    property CycleIcons           : Boolean     read FCycleIcons      write SetCycleIcons;
    property CycleInterval        : Cardinal    read FCycleInterval   write SetCycleInterval;
    property Enabled              : Boolean     read FEnabled         write FEnabled;
    property Hint                 : String      read FHint            write SetHint;
    property ShowHint             : Boolean     read FShowHint        write SetShowHint;
    property Icon                 : TIcon       read FIcon            write SetIcon stored True;
    property IconVisible          : Boolean     read FIconVisible     write SetIconVisible;
    property LeftPopupMenu        : TPopupMenu  read FLeftPopupMenu   write FLeftPopupMenu;
    property RightPopUpMenu       : tPopUpMenu  read FRightPopUpMenu  write FRightPopUpMenu;
  end;


  TScrollingCaption = class(TPersistent)
  private
    FCaption       ,
    FSpace         : String;
    ParentForm     : TForm;
    TmpCount       : Integer;
    FIsMainWindow  : Boolean;
    FDirection     : TScrollDirection;
    FEnabled       : Boolean;
    FInterval      : Word;
    FWindowHandle  : Hwnd;
    procedure UpdateProcessCaption;
    procedure SetCaption(Value: String);
    procedure SetEnabled(Value: Boolean);
    procedure SetInterval(Value: Word);
  protected
    procedure ProcessCaption; dynamic;
  public
    constructor Create;
    destructor Destroy; override;
  published
    property Caption      : String           read FCaption      write SetCaption;
    property Direction    : TScrollDirection read FDirection    write FDirection;
    property IsMainWindow : Boolean          read FIsMainWindow write FIsMainWindow;
    property Space        : String           read FSpace        write FSpace;
    property Enabled      : Boolean          read FEnabled      write SetEnabled   default True;
    property Interval     : Word             read FInterval     write SetInterval default 200;
  end;

  TCover = Class(TPersistent)
  private
   FDrawMethod             : tDrawMethod;
   FClient                 : tBitmap;
   FEnabled                : Boolean;
   procedure SetClient(Value:tBitmap);
   Procedure SetDrawMethod(Value:tDrawMethod);
  protected
  public
   constructor Create;
   destructor Destroy;override;
  published
   property Background       : tBitmap     Read FClient     Write SetClient;
   property BGDrawMethod     : tDrawMethod Read FDrawMethod Write SetDrawMethod;
   property Enabled          : Boolean     Read FEnabled    Write FEnabled;
  end;

  TFormSave = Class(TPersistent)
  private
   FGlobal      : Boolean;
   FKeyName     : String;
   FEnabled     : Boolean;
  protected
  public
   constructor Create;
   destructor Destroy;override;
  published
   property Global   : Boolean Read FGlobal  Write FGlobal;
   property KeyName  : String  Read FKeyname Write FKeyName;
   property Enabled  : Boolean Read FEnabled Write FEnabled;
  end;

  TRatio = class (TPersistent)
  private
   FEnabled              : Boolean;
   FWidth                : Integer;
   FHeight               : Integer;
   FAspectRatio          : Single;
  protected
   procedure SetWidth(value:integer);
   procedure SetHeight(value:integer);
   procedure SetAspectRatio(value:single);
  public
   constructor Create;
   destructor Destroy;override;
  published
   property Enabled       : Boolean read FEnabled     write FEnabled;
   property Width         : Integer read FWidth       write SetWidth;
   property Height        : Integer read FHeight      write SetHeight;
   property AspectRatio   : Single  read FAspectRatio write SetAspectRatio;
  end;

  TResize = class (TPersistent)
  private
   FEnabled      : Boolean;
   FRatio        : TRatio;
   FBorderWidth  : Integer;
   FMaxWidth     : Integer;
   FMinWidth     : Integer;
   FMaxHeight    : Integer;
   FMinHeight    : Integer;
   procedure SetRatio(Value:TRatio);
  public
   constructor Create;
   destructor Destroy;override;
  published
   property Enabled : Boolean     read FEnabled     write FEnabled;
   property Ratio : TRatio        read FRatio       write SetRatio;
   property BorderWidth : Integer read FBorderWidth write FBorderWidth;
   property MaxWidth  : integer   Read FMaxWidth    Write FMaxWidth;
   property MaxHeight : integer   Read FmaxHeight   Write FMaxHeight;
   property MinWidth  : Integer   Read FMinWidth    Write FMinWidth;
   property MinHeight : integer   read FMinHeight   Write FMinHeight;
  end;

  TMargin = class (TPersistent)
  private
   FLeftMin   : integer;
   FLeftMax   : integer;
   FTopMin    : integer;
   FTopMax    : integer;
   FRightMin  : integer;
   FRightMax  : integer;
   FBottomMin : integer;
   FBottomMax : integer;
   FEnabled   : boolean;
  public
   constructor Create;
   destructor Destroy;override;
  published
   property Enabled   : Boolean read FEnabled   write FEnabled;
   property LeftMin : Integer   read FLeftMin   write FLeftMin;
   property LeftMax : Integer   read FLeftMax   write FLeftMax;
   property RightMin : Integer  read FRightMin  write FRightMin;
   property RightMax : Integer  read FRightMax  write FRightMax;
   property TopMin : Integer    read FTopMin    write FTopMin;
   property TopMax : Integer    read FTopMax    write FTopMax;
   property BottomMin : Integer read FBottomMin write FBottomMin;
   property BottomMax : Integer read FBottomMax write FBottomMax;
  end;

  TFormEx = class(TComponent)
  private
   PrevParentWndProc         : Pointer;
   SeekAndDestroy            : Boolean;
   FParent                   : tForm;

   Old                       ,
   Oldh                      ,
   Oldw                      ,
   Oldx                      ,
   Oldy                      : Longint;
   Olds                      : Twindowstate;

   BackGroundBuffer          : tBitmap;

   FShowTaskBar              : Boolean;
   OldStyleEx                : Integer;
   
   FTopMost                  : Boolean;
   FAcceptFiles              : Boolean;
   FBringToFrontOnAccept     : Boolean;
   FAlwaysOnScreen           : Boolean;
   FOnDropFiles              : TDropFilesEvent;
   FOnMovingEvent            : TOnMovingEvent;
   FFullScreen               : Boolean;
   FMargin                   : TMargin;
   FResize                   : TResize;
   FAlwaysMoveable           : Boolean;
   FFormSave                 : tFormSave;
   FCover                    : tCover;
   FScrollingCaption         : tScrollingCaption;
   FSysMenuEx                : tPopUpMenu;
   FTitleBarVisible          : Boolean;

   FTrayIcon                 : tTrayIcon;
   FOnTrayIconClick          ,
   FOnTrayIconDblClick       : TNotifyEvent;
   FOnTrayIconCycle          : TCycleEvent;
   FOnTrayIconMouseDown      ,
   FOnTrayIconMouseUp        : TMouseEvent;
   FOnTrayIconMouseMove      : TMouseMoveEvent;

   FTransparency             : TTransparency;
   
   Procedure SetTitleBarVisible(Value:Boolean);
   Procedure SetSysMenuEx(Value:tPopupMenu);
   Procedure SetTopMost(Value:Boolean);
   Procedure SetShowTaskbar(Value:Boolean);
   procedure SetAcceptFiles(Value: Boolean);
   procedure SetBringToFrontOnAccept(Value: Boolean);
   Procedure SetFullScreen(value:boolean);
   Procedure SetCover(Value:tCover);
  protected
   procedure NewParentWndProc(var Message:Tmessage);
   procedure RebuildBackground;
   procedure DrawBackground(DC:HDC);
   procedure DrawTransparency(DC:HDC);
   procedure TopMostAction;
   Procedure TaskAction;
   Procedure TitleBarAction;
  public
   constructor create(AOwner:TComponent);override;
   destructor  destroy;override;
   procedure   Loaded;override;
   procedure   SaveSettings;
   procedure   LoadSettings;

   Procedure   MinimizeToTray;
   Procedure   RestoreFromTray;
   Procedure   SelfDestroy;
   Procedure   CenterOnForm(Form:tForm);
  published
   property ShowOnTaskBar        : Boolean           read FShowTaskBar          Write SetShowTaskBar;
   property TopMost              : Boolean           read FTopMost              Write SetTopMost;
   property AcceptFiles          : Boolean           read FAcceptFiles          Write SetAcceptFiles;
   property GoToTopOnAcceptFiles : Boolean           read FBringToFrontOnAccept Write SetBringToFrontOnAccept;
   property FullScreen           : Boolean           Read FFulLScreen           Write SetFullScreen;
   property Margin               : TMargin           read FMargin               write FMargin;
   property AlwaysOnScreen       : Boolean           read FAlwaysOnScreen       write FAlwaysOnScreen;
   property AlwaysMoveable       : Boolean           Read FAlwaysMoveable       Write FAlwaysMoveable;
   property Resize               : TResize           read FResize               write FResize;
   property FormSave             : tFormSave         Read FFormSave             Write FFormSave;
   property Cover                : tCover            Read FCover                Write SetCover;
   property ScrollingCaption     : tScrollingCaption Read FScrollingCaption     Write FScrollingCaption;
   property SysMenuEx            : tPopUpMenu        Read FSysMenuEx            Write SetSysMenuEx;
   property TrayIcon             : tTrayIcon     Read FTrayIcon             Write FTrayIcon;
   property ShowTitleBar         : Boolean           Read FTitleBarVisible      Write SetTitleBarVisible;
   property Transparency         : TTransparency     Read FTransparency         Write FTransparency;
   property OnDropFiles          : TDropFilesEvent   read FOnDropFiles          write FOnDropFiles;
   property OnMoving             : TOnMovingEvent    read FOnMovingEvent        write FOnMovingEvent;

   { Events }

   property OnTrayIconClick      : TNotifyEvent      read FOnTrayIconClick      write FOnTrayIconClick;
   property OnTrayIconDblClick   : TNotifyEvent      read FOnTrayIconDblClick   write FOnTrayIconDblClick;
   property OnTrayIconMouseDown  : TMouseEvent       read FOnTrayIconMouseDown  write FOnTrayIconMouseDown;
   property OnTrayIconMouseUp    : TMouseEvent       read FOnTrayIconMouseUp    write FOnTrayIconMouseUp;
   property OnTrayIconMouseMove  : TMouseMoveEvent   read FOnTrayIconMouseMove  write FOnTrayIconMouseMove;
   property OnTrayIconCycle      : TCycleEvent       read FOnTrayIconCycle      write FOnTrayIconCycle;
  end;

procedure Register;

implementation

var IgnoreMessage : Boolean;

type
  PRGBArray = ^TRGBArray;
  TRGBArray = Array[0..1000000] of TRGBTriple;


constructor TTransparency.Create(PForm : tForm);
begin
  inherited Create;

  FTransparency := 60;
  FTranspColor := clBlack;
  ParentForm := PForm;
  ScreenShoot := TBitmap.Create;

  CanTakeScreen := True;
end;

destructor TTransparency.Destroy;
begin
  if ScreenShoot <> nil then ScreenShoot.Destroy;
  inherited Destroy;
end;

procedure TTransparency.TakeScreenShoot;
var
  DC: hDC;
  ScreenWidth, ScreenHeight: Integer;
  SL: PRGBArray;
  X, Y: Integer;
begin
  if Not FEnabled then Exit;
  if Design or  (ScreenShoot = nil) then Exit;

  CanTakeScreen := False;

  ScreenWidth := GetSystemMetrics(sm_CXScreen);
  ScreenHeight := GetSystemMetrics(sm_CYScreen);

  ScreenShoot.Width := ScreenWidth;
  ScreenShoot.Height := ScreenHeight;
  ScreenShoot.PixelFormat := pf24bit;

  IgnoreMessage:=True;
  If ParentForm.Visible then
   ShowWindow(ParentForm.Handle, SW_Hide);

  IgnoreMessage:=True;
  SetActiveWindow(0);

  Sleep(100);
  
  DC := GetDC(0);
  BitBlt(ScreenShoot.Canvas.Handle, 0, 0,
         ScreenWidth, ScreenHeight, DC, 0, 0, SrcCopy);
  ReleaseDC(0, DC);

  for Y := 0 to ScreenHeight - 1 do
   begin
    SL := ScreenShoot.ScanLine[Y];
    for X := 0 to ScreenWidth - 1 do
     begin
      try
       SL[X].rgbtRed := (FTransparency * SL[X].rgbtRed + (100 - FTransparency) * GetRValue(FTranspColor)) div 100;
       SL[X].rgbtGreen := (FTransparency * SL[X].rgbtGreen + (100 - FTransparency)* GetGValue(FTranspColor)) div 100;
       SL[X].rgbtBlue := (FTransparency * SL[X].rgbtBlue + (100 - FTransparency) * GetBValue(FTranspColor)) div 100;
      except
      end;
     end
   end;

  IgnoreMessage:=True;
  If ParentForm.Visible then
  ShowWindow(ParentForm.Handle, sw_Show);
end;

procedure TTransparency.SetTranspColor(Value: TColor);
begin
  if FTranspColor <> Value then
   begin
    FTranspColor := Value;
    if ScreenShoot <> nil then
     begin
      TakeScreenShoot;
     end;
   end;
end;

procedure TTransparency.SetTransparency(Value: THoundred);
begin
  if FTransparency <> Value then
   begin
    FTransparency := Value;
    if ScreenShoot <> nil then
     begin
      TakeScreenShoot;
     end; 
   end;
end;

constructor TTrayIcon.Create(PForm:tForm;PComp:tComponent;InDesign:Boolean);
begin
  inherited Create;
  Design                    := InDesign;                                                          
  FIconVisible              := False;
  FCycleInterval            := 250;
  FEnabled                  := True;
  ParentForm                := PForm;
  ParentComp                := Pcomp;
  
  FIcon                     := TIcon.Create;
  
  IconData.cbSize           := SizeOf(TNotifyIconData);
  IconData.wnd              := AllocateHWnd(HandleIconMessage);
  IconData.uId              := IconID;
  IconData.uFlags           := NIF_ICON + NIF_MESSAGE + NIF_TIP;
  IconData.uCallbackMessage := WM_ICONTRAYNOTIFY;

  ShowIcon;
end;


destructor TTrayIcon.Destroy;
begin
  SetIconVisible(False);
  FIcon.Free;
  DeallocateHWnd(IconData.Wnd);  
  inherited Destroy;
end;

procedure TTrayIcon.Click;
begin
  if Assigned(TFormEx(ParentComp).FOnTrayIconClick) then
    TFormEx(ParentComp).FOnTrayIconClick(Self);
end;


procedure TTrayIcon.DblClick;
begin
  if Assigned(TFormEx(ParentComp).FOnTrayIconDblClick) then
    TFormEx(ParentComp).FOnTrayIconDblClick(Self);
end;


procedure TTrayIcon.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if Assigned(TFormEx(ParentComp).FOnTrayIconMouseDown) then
    TFormEx(ParentComp).FOnTrayIconMouseDown(Self, Button, Shift, X, Y);
end;


procedure TTrayIcon.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if Assigned(TFormEx(ParentComp).FOnTrayIconMouseUp) then
    TFormEx(ParentComp).FOnTrayIconMouseUp(Self, Button, Shift, X, Y);
end;


procedure TTrayIcon.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(TFormEx(ParentComp).FOnTrayIconMouseMove) then
    TFormEx(ParentComp).FOnTrayIconMouseMove(Self, Shift, X, Y);
end;


procedure TTrayIcon.CycleIcon;
begin
  if Assigned(TFormEx(ParentComp).FOnTrayIconCycle) then
    TFormEx(ParentComp).FOnTrayIconCycle(Self, IconIndex);
end;

procedure TTrayIcon.NextIconCycle;
begin
  if Assigned(FIconList) then
  begin
    CycleIcon;
    FIconList.GetIcon(IconIndex, FIcon);
    ModifyIcon;

    if IconIndex < FIconList.Count-1 then  Inc(IconIndex)
    else  IconIndex := 0;
  end;
end;

procedure TTrayIcon.HandleIconMessage(var Msg: TMessage);

function ShiftState: TShiftState;
begin
 Result := [];
 if GetKeyState(VK_SHIFT) < 0 then Include(Result, ssShift);
 if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
 if GetKeyState(VK_MENU) < 0 then Include(Result, ssAlt);
end;

var
  Pt: TPoint;
  Shift: TShiftState;
  I: Integer;
  M: TMenuItem;
begin
  if Msg.Msg = WM_ICONTRAYNOTIFY then
  begin
    case Msg.lParam of

    WM_MOUSEMOVE:
      if FEnabled then
      begin
        Shift := ShiftState;
        GetCursorPos(Pt);
        MouseMove(Shift, Pt.X, Pt.Y);
      end;

    WM_LBUTTONDOWN:
      if FEnabled then
      begin
        Shift := ShiftState + [ssLeft];
        GetCursorPos(Pt);
        MouseDown(mbLeft, Shift, Pt.X, Pt.Y);
        PopUpAtCursor(0);
      end;

    WM_RBUTTONDOWN:
      if FEnabled then
      begin
        Shift := ShiftState + [ssRight];
        GetCursorPos(Pt);
        MouseDown(mbRight, Shift, Pt.X, Pt.Y);
        PopUpAtCursor(1);
      end;

    WM_MBUTTONDOWN:
      if FEnabled then
      begin
        Shift := ShiftState + [ssMiddle];
        MouseDown(mbMiddle, Shift, Pt.X, Pt.Y);
        GetCursorPos(Pt);
      end;

    WM_LBUTTONUP:
      if FEnabled then
      begin
        Shift := ShiftState + [ssLeft];
        GetCursorPos(Pt);
        MouseUp(mbLeft, Shift, Pt.X, Pt.Y);
      end;

    WM_RBUTTONUP:
      if FEnabled then
      begin
        Shift := ShiftState + [ssRight];
        GetCursorPos(Pt);
        MouseUp(mbRight, Shift, Pt.X, Pt.Y);
      end;

    WM_MBUTTONUP:
      if FEnabled then
      begin
        Shift := ShiftState + [ssMiddle];
        GetCursorPos(Pt);
        MouseUp(mbMiddle, Shift, Pt.X, Pt.Y);
      end;

    WM_LBUTTONDBLCLK:
      if FEnabled then
      begin
        DblClick;
        If FLeftPopUpMenu=nil then
        Begin
         M := nil;
         if Assigned(FRightPopupMenu) then
           if (FRightPopupMenu.AutoPopup)  then
             for I := FRightPopUpMenu.Items.Count -1 downto 0 do
             begin
               if FRightPopupMenu.Items[I].Default then
                 M := FRightPopupMenu.Items[I];
             end;
         if M <> nil then
           M.Click;
        End;   
      end;
    end;
  end

  else
    case Msg.Msg of
      WM_QUERYENDSESSION: Msg.Result := 1;
    else      
      Msg.Result := DefWindowProc(IconData.Wnd, Msg.Msg, Msg.wParam, Msg.lParam);
    end;
end;


procedure TTrayIcon.SetIcon(Value: TIcon);
begin
  FIcon.Assign(Value);
  ModifyIcon;
end;

procedure TTrayIcon.SetIconVisible(Value: Boolean);
begin
 if Value then ShowIcon else HideIcon;
end;


procedure TTrayIcon.SetCycleIcons(Value: Boolean);
begin
  If FCycleIcons<>Value then
  Begin
   FCycleIcons := Value;
   If Value then
   Begin
    IconIndex := 0;
//   if (FCycleInterval <> 0) {and FEnabled} then
    if SetTimer(ParentForm.Handle, TrayIconTimerID, FCycleInterval, nil) = 0 then
      raise EOutOfResources.Create('No timers');
   End else
    KillTimer(ParentForm.Handle, TrayIconTimerID);
   End;
end;


procedure TTrayIcon.SetCycleInterval(Value: Cardinal);
begin
  If Value<>FCycleInterval then
  Begin
   FCycleInterval := Value;

   If FCycleIcons then
   Begin
    KillTimer(ParentForm.Handle, TrayIconTimerID);
    if SetTimer(ParentForm.Handle, TrayIconTimerID, FCycleInterval, nil) = 0 then
      raise EOutOfResources.Create('No timers');
   End;
  End;
end;


procedure TTrayIcon.SetHint(Value: String);
begin
 If Value<>FHint then
 Begin
  FHint := Value;
  ModifyIcon;
 End;
end;


procedure TTrayIcon.SetShowHint(Value: Boolean);
begin
 If Value<>FShowHint then
 begin
  FShowHint := Value;
  ModifyIcon;
 end;
end;


function TTrayIcon.InitIcon: Boolean;
begin
  Result := False;
  if Not Design  then
  begin
    IconData.hIcon := FIcon.Handle;
    if (FHint <> '') and (FShowHint) then
      StrLCopy(IconData.szTip, PChar(FHint), SizeOf(IconData.szTip))
    else
      IconData.szTip := '';
    Result := True;
  end;
end;


function TTrayIcon.ShowIcon: Boolean;
begin
  Result := False;
  if Design then Exit;
  if InitIcon then Result := Shell_NotifyIcon(NIM_ADD, @IconData);
end;


function TTrayIcon.HideIcon: Boolean;
begin
 Result := Shell_NotifyIcon(NIM_DELETE, @IconData);
end;


function TTrayIcon.ModifyIcon: Boolean;
begin
 Result := False;
 if InitIcon then
   Result := Shell_NotifyIcon(NIM_MODIFY, @IconData);
end;

procedure TTrayIcon.PopupAtCursor(Index:Integer);
var
  CursorPos: TPoint;
begin
 Case Index of
 0 : Begin
     if Assigned(LeftPopupMenu) then
      if LeftPopupMenu.AutoPopup then
       if GetCursorPos(CursorPos) then
       begin
         Application.ProcessMessages;
         SetForegroundWindow((ParentForm as TWinControl).Handle);
         if Assigned(Screen.ActiveControl) then
           SetFocus(Screen.ActiveControl.Handle);
         LeftPopupMenu.PopupComponent := ParentForm ;
         LeftPopupMenu.Popup(CursorPos.X, CursorPos.Y);
         PostMessage((ParentForm  as TWinControl).Handle, WM_NULL, 0, 0);
       end;
      End;
 1 : Begin
     if Assigned(RightPopupMenu) then
      if RightPopupMenu.AutoPopup then
       if GetCursorPos(CursorPos) then
       begin
         Application.ProcessMessages;
         SetForegroundWindow((ParentForm as TWinControl).Handle);
         if Assigned(Screen.ActiveControl) then
           SetFocus(Screen.ActiveControl.Handle);
         RightPopupMenu.PopupComponent := ParentForm ;
         RightPopupMenu.Popup(CursorPos.X, CursorPos.Y);
         PostMessage((ParentForm  as TWinControl).Handle, WM_NULL, 0, 0);
       end;
      End;
 End;
end;



constructor TScrollingCaption.Create;
begin
  inherited Create;
  FEnabled  := True;
  FInterval := 200;
  FSpace    := '     ';
  TmpCount  := 1;
  UpdateProcessCaption;
end;

destructor TScrollingCaption.Destroy;
begin
  Enabled := False;
  UpdateProcessCaption;
  inherited Destroy;
end;

procedure TScrollingCaption.UpdateProcessCaption;
begin
  KillTimer(FWindowHandle, ScrollCaptionTimerID);
  if (FInterval <> 0) and FEnabled then
    if SetTimer(FWindowHandle, ScrollCaptionTimerID, FInterval, nil) = 0 then
      raise EOutOfResources.Create('No timers');
end;

procedure TScrollingCaption.SetCaption(Value: String);
begin
  if FCaption <> Value then
   begin
    FCaption := Value;
    ParentForm.Caption := Value;
    if FIsMainWindow then
     Application.Title := Value;
    TmpCount := 1;
    UpdateProcessCaption;
   end;
end;

procedure TScrollingCaption.SetEnabled(Value: Boolean);
begin
  if Value <> FEnabled then
  begin
    FEnabled := Value;
    ParentForm.Caption := FCaption;
    if FIsMainWindow then
     Application.Title := FCaption;
    TmpCount := 1;
    UpdateProcessCaption;
  end;
end;

procedure TScrollingCaption.SetInterval(Value: Word);
begin
  if Value <> FInterval then
  begin
    FInterval := Value;
    UpdateProcessCaption;
  end;
end;


procedure TScrollingCaption.ProcessCaption;
var
  St: String;
begin
  try
   St := FCaption + FSpace;
   ParentForm.Caption := Copy(St, TmpCount, Length(St) - TmpCount + 1) + Copy(St, 1, TmpCount - 1);
   if FIsMainWindow then Application.Title := ParentForm.Caption;
   if Direction = dLeft then
    begin
     inc(TmpCount);
     if TmpCount > Length(St) then TmpCount := 1;
    end
   else
    begin
     dec(TmpCount);
     if TmpCount = 0 then TmpCount := Length(St);
    end;
  except
  end;
end;

constructor TCover.Create;
Begin
  inherited Create;
  FClient:=tBitmap.Create;
  FDrawMethod:=dmTile;
End;

destructor TCover.Destroy;
Begin
  FClient.Free;
  inherited Destroy;
End;

procedure TCover.SetClient(Value:tBitmap);
Begin
 If Value<>FClient then
 begin
  FClient.Free;
  FClient:=tBitmap.Create;
  FClient.Assign(Value);
 End;
End;

Procedure tCover.SetDrawMethod(Value:tDrawMethod);
Begin
 if Value<>FDrawMethod then
 Begin
  FDrawMethod:=Value;
 End;
End;

constructor TFormSave.Create;
Begin
  inherited Create;
End;

destructor TFormSave.Destroy;
Begin
  inherited Destroy;
End;

procedure TRatio.SetWidth(value:integer);
begin
 fWidth:=value;
 if Height=0 then fAspectRatio:=0
             else fAspectRatio:=Width/Height;
end;

procedure TRatio.SetHeight(value:integer);
begin
 fHeight:=value;
 if Height=0 then fAspectRatio:=0
             else fAspectRatio:=Width/Height;
end;

procedure TRatio.SetAspectRatio(value:single);
begin
 fAspectRatio:=value;
 fWidth:=100;
 if value=0 then fHeight:=0
            else fHeight:=trunc(100/value);
end;

constructor TRatio.Create;
begin
 inherited Create;
end;

destructor TRatio.Destroy;
begin
 inherited Destroy;
end;

destructor TResize.Destroy;
begin
 fRatio.free;
 inherited Destroy;
end;

constructor TResize.Create;
begin
 inherited Create;
 fRatio:=TRatio.Create;
 fBorderWidth:=5;
end;

procedure TResize.SetRatio(Value:TRatio);
begin
 FRatio.Assign(Value);
end;

constructor TMargin.Create;
begin
 inherited Create;
 LeftMin  :=-5;
 LeftMax  :=10;
 RightMin :=-5;
 RightMax :=10;
 TopMin   :=-5;
 TopMax   :=10;
 BottomMin:=-5;
 BottomMax:=10;
 Enabled:=true;
end;

destructor TMargin.Destroy;
begin
 inherited Destroy;
end;


procedure TFormEx.NewParentWndProc(var Message:Tmessage);
var SkipOldWndProc : Boolean;
    CPoint,
    Pos: TPoint;
    Files : tStrings;
    I,Len,FileCount : Integer;
    FileName: ShortString;
    PR : PRect;
    isLeft,isright,isTop,isbottom:boolean;
    DCHandle : HDC;
Begin
 SkipOldWndProc:=False;
 With Message do
 Begin
   If IgnoreMessage then
   begin
     Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
     Ignoremessage:=False;
     Exit;
   End;

   if (Msg = WM_CLOSE) or (Msg = WM_DESTROY) then
   Begin
     SaveSettings;
     SeekAndDestroy := True;
   End else

   if (Msg = WM_TIMER) then
   Begin
    if not (csDesigning in ComponentState) then
    begin
     if WParam=ScrollCaptionTimerID then
     try
      If FScrollingCaption<>Nil then
       FScrollingCaption.ProcessCaption;
     except
      Application.HandleException(Self);
     end else

     if WParam=TrayIconTimerID then
     try
      If FTrayIcon<>Nil then
       FTrayIcon.NextIconCycle;
     except
      Application.HandleException(Self);
     end

    End;
   End else
 
   if (Msg=WM_ERASEBKGND) then
   Begin
    If FCover.Enabled then
    Begin
//     Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
     RebuildBackground;
     DCHandle:= GetDC(FParent.Handle);
     DrawBackground(DCHandle);
     ReleaseDC(FParent.Handle,DCHandle);
     SkipOldWndProc:=True;
//   Result:=1;
    End else
    If FTransparency.Enabled then
    Begin
//     Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
     DCHandle:= GetDC(FParent.Handle);
     DrawTransparency(DCHAndle);
     ReleaseDC(FParent.Handle,DCHandle);
     SkipOldWndProc:=True;
     Result:=1;
    End;
   End else

   If (Msg=WM_MOVE) then
   Begin
    If FTransparency.Enabled then
    Begin
     Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
     DCHandle:= GetDC(FParent.Handle);
     DrawTransparency(DCHAndle);
     ReleaseDC(FParent.Handle,DCHandle);
     SkipOldWndProc:=True;
    End;
   End;

{   If (Msg=WM_SIZE) then
   Begin
    If wParam = SIZE_MINIMIZED then
    begin
     if FMinimizeToTray then DoMinimizeToTray;
    End;
   End;
}
   If (Msg=WM_ACTIVATE) then
   Begin
    if (Lo(wParam) <> wa_Inactive) and
     FTransparency.CanTakeScreen then FTransparency.TakeScreenShoot;
   End;

   If (Msg=WM_MOVING) then
   Begin
    PR := Pointer(LParam);

    If Assigned (OnMoving) then
    Begin
     OnMoving(self,pr^);
     Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
     Exit;
    end;

    If ((PR^.left < Margin.LeftMax) and (PR^.Left>Margin.LeftMin) and (Margin.Enabled)) or
       ((AlwaysOnScreen) and (PR^.Left<0))then
     begin
      PR^.left := 0;
      PR^.Right := FParent.width;
     End;

    if ((PR^.top < Margin.TopMax) and (PR^.Top>Margin.TopMin) and (Margin.Enabled)) or
       ((AlwaysOnScreen) and (PR^.Top<0)) then
     begin
      PR^.top := 0;
      PR^.Bottom := FParent.height;
     end;

    if ((PR^.Bottom > screen.Height-Margin.BottomMax) and
       (PR^.Bottom+Margin.BottomMin < screen.Height) and (Margin.Enabled)) or
       ((AlwaysOnScreen) and (PR^.Bottom>screen.height)) then
     begin
      PR^.Bottom := screen.height;
      PR^.top := screen.height - FParent.height;
     end;

    if ((PR^.Right > screen.Width-Margin.RightMax) and
       (PR^.Right + Margin.RightMin< screen.Width) and (Margin.Enabled)) or
       ((AlwaysOnScreen) and (PR^.Right>Screen.Width)) then
     begin
      PR^.Right := screen.width;
      PR^.Left := screen.width - FParent.width;
     end;
     Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
   End else

   If (Msg=WM_NCHITTEST) then
   Begin
    Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
    SkipOldWndProc:=True;
    If Resize.Enabled then
    begin
     Pos.x:=LoWord(LParam);
     Pos.y:=HiWord(LParam);

     CPoint:=FParent.ScreenToClient(Pos);

     isLeft  :=(CPoint.x<Resize.BorderWidth);
//     isTop   :=(CPoint.y<Resize.BorderWidth);
     { Assumed all forms has a caption }
     isTop   :=Pos.Y<FParent.Top+Resize.BorderWidth;
     isRight :=(CPoint.x+Resize.BorderWidth>=fParent.ClientWidth);
     isBottom:=(CPoint.y+Resize.BorderWidth>=fParent.ClientHeight);

     if IsLeft then
      if isTop    then Result:=HTTOPLEFT else
      if isBottom then Result:=HTBOTTOMLEFT else
                       Result:=HTLEFT
     else

     if IsRight then
      if isTop then    Result:=HTTOPRIGHT else
      if isBottom then Result:=HTBOTTOMRIGHT else
                       Result:=HTRIGHT
     else

     if IsTop then
      Result:=HTTOP

     else

     if IsBottom then
      Result:=HTBOTTOM;
      
{     else
      Result:=HTCLIENT;
}   end;

    If (Result = htClient) and (FAlwaysMoveable) then
    Begin

     Pos.x:=LoWord(LParam);
     Pos.y:=HiWord(LParam);

     Pos:=FParent.ScreenToClient(Pos);
     Result:=htCaption;
     if (FParent is TForm) then
      with (FParent as TForm) do
      begin
       for i := 0 to ComponentCount - 1 do
       Begin
        if Components[i] is TGraphicControl then
        Begin
        With (Components[i] as TGraphicControl) do
         Begin
         If (Pos.X >= Left) and (Pos.X<=Left+Width) and
            (Pos.Y >= Top ) and (Pos.Y<=Top+Height) and
            (Align=alNone) then
         Begin
          Result:=htClient;
          Break;
         End;
        End;
       End;
      end;
     End;

    End;
   End else

   If (Msg=WM_SIZING) then
   Begin
    if (not(Resize.Ratio.Enabled)) or (Resize.Ratio.AspectRatio=0) then exit;
    PR := Pointer(LParam);
    if WParam=WMSZ_BOTTOMRIGHT then PR^.right:=PR^.Left+trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio) else
    if WParam=WMSZ_BOTTOMLEFT  then PR^.left :=PR^.Right-trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio) else
    if WParam=WMSZ_TOPLEFT     then PR^.left :=PR^.Right-trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio) else
    if WParam=WMSZ_TOPRIGHT    then PR^.right:=PR^.Left+trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio);
    SkipOldWndProc:=True;
   End else

   If (Msg=WM_GETMINMAXINFO) then
   Begin
    Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
    With PMinMaxInfo(lParam)^ do begin
    With FResize do
    Begin
     if (FMaxWidth <> 0) then ptMaxTrackSize.X := FMaxWidth;
     if (FMaxHeight <> 0) then ptMaxTrackSize.Y := FMaxHeight;
     if (FMinWidth <> 0) then ptMinTrackSize.X := FMinWidth;
     if (FMinHeight <> 0) then ptMinTrackSize.Y := FMinHeight;
    End; end;
   End else

   If (Msg=WM_SYSCOMMAND) then
   Begin
    If (WParam>SysMenuExID) and (FSysMenuEx<>nil) then
    Begin
     For I:=0 to FSysMenuEx.Items.Count-1 do
     Begin
      If FSysMenuEx.Items[I].Tag=WParam-SysMenuExID then
       FSysMenuEx.Items[I].Click;
     End;
    End;
    if (WParam=SC_MINIMIZE) then
    Begin
     ShowWindow(FParent.Handle,SW_MINIMIZE);
    End else
    If (WParam=SC_RESTORE) then
    Begin
     ShowWindow(FParent.Handle,SW_RESTORE);
    End;
   End else
   
   If (Msg=WM_SHOWWINDOW) then
   Begin
    if Bool(Wparam) then
    Begin
     TaskAction;
     LoadSettings;
    End else
    Begin
     SkipOldWndProc:=True;
     ShowWindow(FParent.Handle,SW_MINIMIZE);
     TaskAction;
     SaveSettings;
    End;
   end else

  If (Msg=WM_DROPFILES) then
  Begin
   if not (csDesigning in ComponentState) then begin
    if FBringToFrontOnAccept then
     SetForegroundWindow(FParent.Handle);

    DragQueryPoint(wParam, Pos);
    Files := TStringList.Create;
    try
      FileCount := DragQueryFile(wParam, UINT(-1), nil, 0);
      for I := 0 to (FileCount - 1) do begin
        Len := DragQueryFile(wParam, I, @FileName[1], 255);
        FileName[0] := Char(Len);
        Files.Add(FileName);
      end;
      if (FileCount > 0) and Assigned(FOnDropFiles) then
        FOnDropFiles(Files, Pos);
    finally
     Files.Free;
    end;
   End;
  end;

  If Not SkipOldWndProc then
    Result := CallWindowProc(PrevParentWndProc, FParent.Handle, Msg, WParam, LParam);
 End;
End;

Constructor TFormEx.create(AOwner:TComponent);
Var P : Pointer;
Begin
 inherited Create(AOwner);
 FParent:=TForm(AOwner);

 BackgroundBuffer:=tBitmap.Create;

 If Not (csDesigning in ComponentState) then
 Begin
  PrevParentWndProc := Pointer(GetWindowLong(FParent.Handle, GWL_WNDPROC));
  P := MakeObjectInstance(NewParentWndProc);
  SetWindowLong(FParent.Handle, GWL_WNDPROC, LongInt(p));
 End;

 FMargin:=TMargin.Create;
 FResize:=TResize.Create;
 FResize.Ratio.Width:=FParent.Width;
 FResize.Ratio.Height:=FParent.Height;
 FCover:=tCover.Create;
 { Default Value }
 FShowTaskBar:=False;
 OldStyleEx:=GetWindowLong(FParent.Handle,GWL_EXSTYLE);
 FTopMost:=False;
 FAcceptFiles:=False;
 FBringToFrontOnAccept:=False;
 FFulLScreen:=False;
 FTitleBarVisible:=true;

 // If FParent.BorderStyle=bsNone then FAlwaysMoveable:=true;

 FFormSave:=tFormSave.Create;
 With FFormSave do
 Begin
  FGlobal  :=True;
  FKeyName :='Software\'+Application.Title+'\'+FParent.Caption;
  FEnabled :=True;
 End;

 FScrollingCaption:=tScrollingCaption.Create;
 FScrollingCaption.ParentForm:=FParent;
 FScrollingCaption.FWindowHandle:=FParent.Handle;

 FTrayIcon:=tTrayIcon.Create(FParent,Self,(csDesigning in ComponentState));

 FTransparency:=TTransparency.Create(FParent);
 FTransparency.Design:=(csDesigning in ComponentState);
End;

procedure TFormEx.Loaded;
begin
 inherited Loaded;
 if not (csDesigning in ComponentState) then
 Begin
  LoadSettings;
  TaskAction;
  TopMostAction;
  If FTrayIcon<>nil then
  With FTrayIcon do
  Begin
   ModifyIcon;
   SetIconVisible(FIconVisible);
  End;

  If FTransparency<>nil then FTransparency.TakeScreenShoot;
 End;
end;

Destructor TFormEx.destroy;
Begin
 FTransparency.Free;
 FTrayIcon.Free;
 FScrollingCaption.Free;

 FMargin.Free;
 FResize.Free;
 FFormSave.Free;
 FCover.Free;
 BackgroundBuffer.Free;
 If Not (csDesigning in ComponentState) then
 Begin
  if not SeekAndDestroy then
   begin
    SetWindowLong(FParent.Handle, GWL_WNDPROC, LongInt(PrevParentWndProc));
   end;
 End;
 inherited destroy;
End;


Procedure TFormEx.TaskAction;
begin
 if FShowTaskBar then
 Begin
  If (OldStyleEX and WS_EX_APPWINDOW)<>WS_EX_APPWINDOW then
    SetWindowLong(FParent.Handle,GWL_EXSTYLE,OldStyleEX or (WS_EX_APPWINDOW or WS_EX_CONTROLPARENT))
 End else
  SetWindowLong(FParent.Handle,GWL_EXSTYLE,OldStyleEX);
 { Warning, above routine replace the Window EX Style }
  if not (csDesigning in ComponentState) then
   DragAcceptFiles(FParent.Handle, FAcceptFiles)
End;

Procedure TFormEx.SetShowTaskBar(Value:Boolean);
Begin
 If Value<>FShowTaskBar then
 Begin
  FShowTaskBar:=Value;
 If Not (csDesigning in ComponentState) then
  TaskAction;
 End;
End;

Procedure TFormEx.TopMostAction;
Begin
 If FTopMost then
 SetWindowPos(FParent.Handle, HWND_TOPMOST, 0,0,0,0, SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE) else
 SetWindowPos(FParent.Handle, HWND_NOTOPMOST,0,0,0,0, SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE);
End;

Procedure TFormEx.SetTopMost(Value:Boolean);
begin
 if Value<>FTopMost then
 Begin
  FTopMost := Value;
 If Not (csDesigning in ComponentState) then
  TopMostAction;
 End;
end;

procedure TFormEx.SetAcceptFiles(Value: Boolean);
begin
 if (Value <> FAcceptFiles) then begin
   FAcceptFiles := Value;
   if not (csDesigning in ComponentState) then
     DragAcceptFiles(FParent.Handle, FAcceptFiles)
 end;
end;

procedure TFormEx.SetBringToFrontOnAccept(Value: Boolean);
begin
 if (Value <> FBringToFrontOnAccept) then begin
   FBringToFrontOnAccept := Value;
 end;
end;

Procedure TFormEx.SetFullScreen(value:boolean);
Begin
 If Value<>Ffullscreen Then
 Begin
  Ffullscreen:=Value;
  If Ffullscreen Then
  Begin
    if not (csDesigning in ComponentState) then begin
    Old:=Getwindowlong(Fparent.Handle, Gwl_Style);
    Setwindowlong(Fparent.Handle, Gwl_Style, Getwindowlong(Fparent.Handle, Gwl_Style) And Not Ws_Caption);
    Oldh:=Fparent.Height;
    Oldw:=Fparent.Width;
    Oldx:=Fparent.Left;
    Oldy:=Fparent.Top;
    Olds:=Fparent.Windowstate;
    Fparent.Windowstate:=Wsmaximized;
    Fparent.Clientheight:=Screen.Height;
    Fparent.Refresh;
    End;
  End
  Else
  Begin
    if not (csDesigning in ComponentState) then begin
    Setwindowlong(Fparent.Handle, Gwl_Style, Old);
    Fparent.Height:=Oldh;
    Fparent.Width:=Oldw;
    Fparent.Left:=Oldx;
    Fparent.Top:=Oldy;
    Fparent.Windowstate:=Olds;
    Fparent.Refresh;
    End;
  End;
 End;
End;

procedure TFormEx.SaveSettings;
Begin
 If Not FFormSave.FEnabled then Exit;

 With TRegistry.Create do
 Begin
   If FFormSave.FGlobal then
   RootKey:=HKEY_LOCAL_MACHINE else
   RootKey:=HKEY_CURRENT_USER;
   If OpenKey(FFormSave.FKeyName,true) then
   Begin
     WriteInteger('Width',FParent.Width);
     WriteInteger('Height',FParent.Height);
     WriteInteger('X',FParent.Left);
     WriteInteger('Y',FParent.Top);
     WriteBool('FullScreen',FFullScreen);
   End;

 End;

End;

procedure TFormEx.LoadSettings;
Begin
 If Not FFormSave.FEnabled then Exit;

 With TRegistry.Create do
 Begin

   If FFormSave.FGlobal then
   RootKey:=HKEY_LOCAL_MACHINE else
   RootKey:=HKEY_CURRENT_USER;
   If OpenKey(FFormSave.FKeyName,False) then
   Begin
     If ValueExists('Width') then
      FParent.Width:=ReadInteger('Width');

     If ValueExists('Height') then
      FParent.Height:=ReadInteger('Height');

     If ValueExists('X') then
      FParent.Left:=ReadInteger('X');

     If ValueExists('Y') then
      FParent.Top:=ReadInteger('Y');

     If ValueExists('FulLScreen') then
      FullScreen:=ReadBool('FullScreen');
   End;

 End;
End;

Procedure tFormEx.SetCover(Value:tCover);
Begin
 If Value<>FCover then
 begin
  FCover:=Value;
 End;
End;

Procedure tFormEx.RebuildBackground;
var w,h,x,y:integer;
    Buffer:tBitmap;
    Width,Height : Integer;
Begin
 If FCover.Background.Empty then
 Begin
  Exit;
 End;

 Width:=FParent.ClientWidth;
 Height:=FParent.ClientHeight;

 With FCover do
 Begin

  Buffer:=tbitmap.create;
  Buffer.width:=Width;
  Buffer.Height:=Height;

  If FDrawMethod=dmNormal then
  Begin
   BitBlt(Buffer.Canvas.Handle,0,0,FCover.Background.Width,FCover.Background.Height,FCover.Background.Canvas.Handle,0,0,SRCCopy);
  End else

  If fDrawMethod=dmCenter then
  begin
   x:=(Width-FCover.Background.Width) div 2;
   y:=(Height-FCover.Background.Height) div 2;
   BitBlt(Buffer.Canvas.Handle,X,Y,FCover.Background.Width,FCover.Background.Height,FCover.Background.Canvas.Handle,0,0,SRCCopy);
  End else
 
  If FDrawMethod=dmStretch then
  Begin
   StretchBlt(Buffer.Canvas.Handle,0,0,Width,Height,FCover.Background.Canvas.Handle,0,0,FCover.Background.Width,FCover.Background.Height,SRCCopy);
  End else

  If FDrawMethod=dmTile then
  Begin
   X:=0;
   Y:=0;
   W:=FCover.Background.Width;
   H:=FCover.Background.Height;

   While (X < Width) do
   Begin
    Y:=0;
    while (Y < Height) do
    Begin
     BitBlt(Buffer.Canvas.Handle,X,Y,FCover.Background.Width,FCover.Background.Height,FCover.Background.Canvas.Handle,0,0,SRCCopy);
     Inc(Y,H);
    end;
    Inc(X,W);
   End;
  End;

  BackgroundBuffer.Free;
  BackgroundBuffer:=tBitmap.Create;
  BackgroundBuffer.Assign(Buffer);

  Buffer.Free;
 End;
End;

Procedure tFormEx.DrawTransparency(DC:HDC);
Begin
 If FTransparency=nil then Exit;
 if Not FTransparency.FEnabled then Exit;
 FTransparency.CanTakeScreen := False;
 if FTransparency.ScreenShoot <> nil then
  BitBlt(DC, 0, 0, FTransparency.ScreenShoot.Width, FTransparency.ScreenShoot.Height,
         FTransparency.ScreenShoot.Canvas.Handle, FParent.Left, FParent.Top, SrcCopy);
 FTransparency.CanTakeScreen := True;
End;

procedure tFormEx.DrawBackground(DC:HDC);
var Width,Height : Integer;
Begin
 Width:=FParent.ClientWidth;
 Height:=FParent.ClientHeight;

 BitBlt(DC,0,0,Width,Height,BackgroundBuffer.Canvas.Handle,0,0,SRCCopy);
End;

Procedure tFormEx.SetSysMenuEx(Value:tPopupMenu);
var SysMenu : HMenu;
    Count   : Integer;
Begin
 If Value<>FSysMenuEx then
 Begin
  SysMenu:=GetSystemMenu(FParent.Handle,True);

  SysMenu:=GetSystemMenu(FParent.Handle,False);
  FSysMenuEx:=Value;
  If FSysMenuEx=nil then Exit;
  For count:=0 to FSysMenuEx.Items.Count-1 do
  Begin
   If FSysMenuEx.Items[Count].Caption<>'-' then
   AppendMenu(SysMenu, mf_ByCommand, FSysMenuEx.Items[Count].Tag + SysMenuExID, Pchar(FSysMenuEx.Items[Count].Caption)) else
   AppendMenu(SysMenu, mf_ByCommand or MF_SEPARATOR, 0, '');
  End;
 End;
End;


procedure tFormEx.SetTitlebarVisible(value: boolean);
begin
 If Value<>FTitleBarVisible then
 Begin
  FTitleBarVisible := Value;
  If Not (csDesigning in ComponentState) then
   TitleBarAction;
 End;
end;

Procedure tFormEx.TitleBarAction;
Var
  Save : LongInt;
Begin
  if FParent = nil then exit;
  with FParent do
  begin
      case BorderStyle of
        bsNone,
        bsSizeToolWin,
        bsToolWindow: Exit;
      end;
      Save:=GetWindowLong(Handle,gwl_Style);
      If (Save and ws_Caption)=ws_Caption then Begin
        Case BorderStyle of
          bsSingle,
          bsSizeable : SetWindowLong(Handle,gwl_Style,Save and
            (Not(ws_Caption)) or ws_border);
          bsDialog : SetWindowLong(Handle,gwl_Style,Save and
            (Not(ws_Caption)) or ds_modalframe or ws_dlgframe);
        End;

      if Not FTitleBarVisible then
      begin
        Height:=Height+getSystemMetrics(sm_cyCaption);
      end else
        Height:=Height-getSystemMetrics(sm_cyCaption);

      if FTitleBarVisible then
      begin
        Height:=Height-getSystemMetrics(sm_cyCaption);
      end else
        Height:=Height+getSystemMetrics(sm_cyCaption);

      Refresh;
    End;
  End;
end;


Procedure TFormEx.MinimizeToTray;
Begin
 FParent.Hide;
 ShowWindow(FParent.Handle,SW_Minimize);
End;

Procedure TFormEx.RestoreFromTray;
Begin
 FParent.Show;
 ShowWindow(FParent.Handle,SW_RESTORE);
End;

Procedure TFormEx.SelfDestroy;
Begin
 FParent.Release;
End;

Procedure TFormEx.CenterOnForm(Form:tForm);
Begin
 Fparent.Left:=(Form.ClientWidth-FParent.Width) div 2;
 FParent.Top :=(Form.ClientHeight-FParent.Height) div 2;
End;

procedure Register;
begin
  RegisterComponents('Jazarsoft', [TFormEx]);
end;

end.
