unit MouseHook;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TMouseProcEvent = function (Sender: TObject; nCode : Integer; wParam : Word; lParam : longint) : longint of object;
  TMouseHook = class(TComponent)
  private
    FOnMouseProc : TMouseProcEvent;
    FMouseHookStruct : TMouseHookStruct;
    FHooked : Boolean;
    procedure SetHooked (Value : Boolean);
    function GetHooked : Boolean;
  protected
    function DoMouseProc (nCode : Integer; wParam : Word; lParam : Longint) : longint; virtual;
  public
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
    procedure ReInstallHook;
    procedure InstallHook; virtual;
    procedure UninstallHook; virtual;
    property MouseHookStruct : TMouseHookStruct read FMouseHookStruct;
  published
    property Hooked : Boolean read GetHooked Write SetHooked default false;
    property OnMouseProc : TMouseProcEvent read FOnMouseProc Write FOnMouseProc;
  end;

procedure Register;

implementation

var
  MouseHooks : TList;
  HHook : THandle;

{ TMouseHook }

function MouseProc (nCode : integer; wParam : integer; lParam : Longint) : longint; stdcall;
var
  i : Integer;
begin
  Result := 0;
  for i := 0 to MouseHooks.Count - 1 do
    begin
      Result := TMouseHook (MouseHooks [i]).DoMouseProc (nCode, wParam, lParam);
      if Result <> 0
        then Break;
    end;
  if Result = 0
    then Result := CallNextHookEx (HHook, nCode, wParam, lParam);
end;

constructor TMouseHook.Create;
begin
  inherited;
  MouseHooks.Add (Self);
end;

destructor TMouseHook.Destroy;
begin
  MouseHooks.Remove (Self);
  if Hooked and (MouseHooks.Count <= 0)
    then if not (csDesigning in ComponentState)
      then UninstallHook;
  inherited;
end;

function TMouseHook.DoMouseProc;
begin
  if Assigned (FOnMouseProc)
    then
    begin
      FMouseHookStruct := PMouseHookStruct (Pointer (lParam))^;
      Result := FOnMouseProc (self, nCode, wParam, lParam);
    end
    else Result := 0;
end;

procedure TMouseHook.ReInstallHook;
begin
  if Hooked
    then
    begin
      UninstallHook;
      InstallHook;
    end;
end;

procedure TMouseHook.InstallHook;
begin
  if HHook = 0
    then
    begin
      HHook := SetWindowsHookEx (WH_MOUSE, @MouseProc, 0, GetCurrentThreadId);
      FHooked := HHook <> 0;
    end;
end;

procedure TMouseHook.UninstallHook;
begin
  if HHook <> 0
    then
    begin
      UnhookWindowsHookEx (HHook);
      HHook := 0;
      FHooked := False;
    end;
end;

procedure TMouseHook.SetHooked;
begin
  if Hooked <> Value
    then if not (csDesigning in ComponentState)
      then if value
        then InstallHook
        else UninstallHook
      else FHooked := Value;
end;

function TMouseHook.GetHooked;
begin
  if csDesigning in ComponentState
    then Result := FHooked
    else Result := HHook <> 0;
end;

{ Register Proc }

procedure Register;
begin
  RegisterComponents('Misc2', [TMouseHook]);
end;

initialization
  MouseHooks := TList.Create;
finalization
  MouseHooks.Free;
end.
