(*
  TSparseList -> TList
  A simple descendant of TList that reduces memory fragmentation.  When items
  are deleted from the list, the slots are kept open so newly added items can
  be added into those slots.  This avoids having to move memory whenever an item
  is deleted, and allows the list to reclaim memory used by deleted items.

  Use only the Add and Remove methods, or the Clear method to clear the list.
  The Insert method should not be used.
*)

unit SparseList;

interface

uses
  Windows, Messages, SysUtils, Classes;

type

  TSparseList = class( TList )
  private
    nDeleted: integer;
  protected
  public
    constructor Create;
    function Add( Item: pointer ): integer; reintroduce;
    procedure Clear; override;
    procedure Delete( Index: integer ); reintroduce;
  end;

implementation

{ TSparseList }

function TSparseList.Add(Item: pointer): integer;
begin
  if nDeleted = 0 then
    Result := inherited Add( Item )
  else
  begin
    Result := 0;
    while Items[Result] <> nil do
      Inc( Result );
    Items[Result] := Item;
    Dec( nDeleted );
  end;
end;

procedure TSparseList.Clear;
begin
  inherited Clear;
  nDeleted := 0;
end;

constructor TSparseList.Create;
begin
  nDeleted := 0;
end;

procedure TSparseList.Delete(Index: integer);
begin
  Items[Index] := nil;
  Inc( nDeleted );
end;

end.
