//******************************************************************************
//* TStringGridLink                                                            *
//*                                                                            *
//* version:     1.0                                                           *
//* date:        12.08.2000                                                    *
//* compiler:    Delphi 5.0                                                    *
//* author:      Inger Hoffmann copyright  2000                               *
//* e-mail:      ingerhoffmann@web.de                                          *
//*                                                                            *
//******************************************************************************
//* important note:                                                            *
//* The source code is given as is. The author is not responsible              *
//* for any possible damage done due to the use of this code.                  *
//* The component can be freely used in any application. The source            *
//* code remains property of the author and may not be distributed             *
//* freely as such nor sold in any form.                                       *
//******************************************************************************
//* description:                                                               *
//* TStringGridLink is a non visual Component, that is linking two             *
//* TStringGrid controls for moving single rows from one StringGrid            *
//* to the other StringGrid and back.                                          *
//******************************************************************************
unit StringGridLink;

interface

uses
  SysUtils, Classes, Controls, stdctrls, grids;

type
  TClickEvent = procedure (Sender:TObject) of object;
  TStringGridLink = class(TComponent)
  private
   FMoveFromTo: TButton;
   FMoveToFrom: TButton;
   FStringGridFrom: TStringGrid;
   FStringGridTo: TStringGrid;
   FStringGridFromFixedRows: integer;
   FStringGridToFixedRows: integer;
   FMoveFromToClick: TClickEvent;
   FMoveToFromClick: TClickEvent;
   procedure SetMoveFromTo(abutton: TButton);
   procedure SetMoveToFrom(abutton: TButton);
   procedure SetStringGridFrom(aStringGrid: TStringGrid);
   procedure SetStringGridTo(aStringGrid: TStringGrid);
   procedure MoveFromToClick(Sender: TObject);
   procedure MoveToFromClick(Sender: TObject);
   procedure MoveStringGridData(SourceStringGrid,TargetStringGrid: TStringGrid;
                                SourceStringGridFixedRows,TargetStringGridFixedRows : integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
  published
    property MoveFromTo: TButton read FMoveFromTo write SetMoveFromTo;
    property MoveToFrom: TButton read FMoveToFrom write SetMoveToFrom;
    property StringGridFrom: TStringGrid read FStringGridFrom write SetStringGridFrom;
    property StringGridTo: TStringGrid read FStringGridTo write SetStringGridTo;
  end;

procedure Register;

implementation

//******************************************************************************
constructor TStringGridLink.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

//******************************************************************************
destructor TStringGridLink.Destroy;
begin
  inherited Destroy;
end;

//******************************************************************************
procedure TStringGridLink.SetStringGridFrom(aStringGrid:TStringGrid);
begin
  if not assigned(aStringGrid) then exit;
  FStringGridFrom := aStringGrid;
  if FStringGridFrom.FixedRows > 0 then
    FStringGridFromFixedRows := FStringGridFrom.FixedRows
  else
    FStringGridFromFixedRows := 0;
end;

//******************************************************************************
procedure TStringGridLink.SetStringGridTo(aStringGrid:TStringGrid);
begin
  if not assigned(aStringGrid) then exit;
  FStringGridTo := aStringGrid;
  if FStringGridTo.FixedRows > 0 then
    FStringGridToFixedRows := FStringGridTo.FixedRows
  else
    FStringGridToFixedRows := 0;
end;

//******************************************************************************
procedure TStringGridLink.SetMoveFromTo(abutton:TButton);
begin
  if not assigned(abutton) then exit;
  FMoveFromTo := abutton;
  if assigned(FMoveFromTo.OnClick) then
    FMoveFromToClick := FMoveFromTo.OnClick;
  FMoveFromTo.OnClick := MoveFromToClick;
end;

//******************************************************************************
procedure TStringGridLink.SetMoveToFrom(abutton:tbutton);
begin
  if not assigned(abutton) then exit;
  FMoveToFrom := abutton;
  if assigned(FMoveToFrom.OnClick) then
    FMoveToFromClick := FMoveToFrom.OnClick;
  FMoveToFrom.OnClick := MoveToFromClick;
end;

//******************************************************************************
procedure TStringGridLink.MoveStringGridData;
var
  SourceRowList, TargetRowList : TStringList;
  i, iCount, iSourceIndex : integer;
  bSourceRowEmpty, bTargetRowEmpty, SourceStringGridLastRow : boolean;
begin
  try
    SourceRowList := TStringList.Create;
    TargetRowList := TStringList.Create;
    try
      if not (SourceStringGridFixedRows = SourceStringGrid.RowCount) then
      begin

        // Source
        SourceRowList.Clear;
        SourceRowList.AddStrings(SourceStringGrid.Rows[SourceStringGrid.Row]);
        iSourceIndex := SourceStringGrid.Row;
        for i := iSourceIndex to SourceStringGrid.RowCount - 1 do
        begin
          SourceStringGrid.Rows[i].Clear;
          SourceStringGrid.Rows[i].AddStrings(SourceStringGrid.Rows[i+1]);
        end;
        SourceStringGridLastRow := true;
        bSourceRowEmpty := false;
        if (SourceStringGrid.RowCount = 1) and (SourceStringGridFixedRows = 0) then
        begin
          bSourceRowEmpty := true;
          for i := 0 to SourceRowList.Count - 1 do
          begin
            if not ((SourceRowList.Strings[i] = EmptyStr) or
              (SourceRowList.Strings[i] = ' ')) then bSourceRowEmpty := false;
          end;
        end
        else
          SourceStringGridLastRow := false;
        SourceStringGrid.RowCount := SourceStringGrid.RowCount - 1;

        // Target
        iCount := TargetStringGrid.RowCount;
        if not (SourceStringGridLastRow and bSourceRowEmpty) then
        begin
          TargetRowList.AddStrings(TargetStringGrid.Rows[TargetStringGrid.RowCount-1]);
          bTargetRowEmpty := true;
          for i := 0 to TargetRowList.Count - 1 do
          begin
            if not ((TargetRowList.Strings[i] = EmptyStr) or
              (TargetRowList.Strings[i] = ' ')) then bTargetRowEmpty := false;
          end;
          if not ((TargetStringGrid.RowCount = 1) and
            (TargetStringGridFixedRows = 0) and
            (bTargetRowEmpty = true)) then
          begin
            TargetStringGrid.RowCount := TargetStringGrid.RowCount + 1;
            TargetStringGrid.Rows[iCount].Clear;
            TargetStringGrid.Rows[iCount].AddStrings(SourceRowList);
            if (iCount = TargetStringGridFixedRows) then
              TargetStringGrid.FixedRows := TargetStringGridFixedRows;
          end
          else
          begin
            TargetStringGrid.Rows[iCount-1].Clear;
            TargetStringGrid.Rows[iCount-1].AddStrings(SourceRowList);
            if (iCount = TargetStringGridFixedRows) then
              TargetStringGrid.FixedRows := TargetStringGridFixedRows;
          end;
        end;
      end;
    finally
      SourceRowList.Free;
      TargetRowList.Free;
    end;
  except
    exit;
  end;
end;

//******************************************************************************
procedure TStringGridLink.MoveToFromClick(sender:tobject);
begin
  if assigned(FMoveToFromClick) then
  begin
    try
      FMoveToFromClick(Sender);
    except
      exit;
    end;
  end;
  if not assigned(FStringGridFrom) then exit;
  if not assigned(FStringGridTo) then exit;
  if (FStringGridTo.RowCount = 0) then exit;
  MoveStringGridData(FStringGridTo,FStringGridFrom,FStringGridToFixedRows,FStringGridFromFixedRows);
end;

//******************************************************************************
procedure TStringGridLink.MoveFromToClick(sender:tobject);
begin
  if assigned(FMoveFromToClick) then
  begin
    try
      FMoveFromToClick(Sender);
    except
      exit;
    end;
  end;
  if not assigned(FStringGridFrom) then exit;
  if not assigned(FStringGridTo) then exit;
  if (FStringGridFrom.RowCount = 0) then exit;
  MoveStringGridData(FStringGridFrom,FStringGridTo,FStringGridFromFixedRows,FStringGridToFixedRows);
end;

//******************************************************************************
procedure Register;
begin
  RegisterComponents('External', [TStringGridLink]);
end;

end.
