unit WhatIsThis;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, MouseHook;

type
  TWhatIsThisEvent = procedure (Sender : TObject; Control : TControl; HelpContext : THelpContext; var ShowHelp : boolean) of object;
  TNoHelpContextEvent = procedure (Sender : TObject; Control : TControl) of object;
  TWhatIsThis = class(TComponent)
  private
    FMouseHook : TMouseHook;
    FActive : Boolean;
    FOnWhatIsThis : TWhatIsThisEvent;
    FOnNoHelpContext : TNoHelpContextEvent;
    FOnActiveChanged : TNotifyEvent;
    function MouseHookMouseProc(Sender: TObject; nCode: Integer;
                                wParam: Word; lParam: Integer): Integer;
    procedure SetActive (Value : Boolean);
  public
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
    property Active : Boolean read FActive Write SetActive;
  published
    property OnWhatIsThis : TWhatIsThisEvent read FOnWhatIsThis Write FOnWhatIsthis;
    property OnNoHelpContext : TNoHelpContextEvent read FOnNoHelpContext Write FOnNoHelpContext;
    property OnActiveChanged : TNotifyEvent read FOnActiveChanged Write FOnActiveChanged;
  end;

procedure Register;

implementation

uses
  typInfo;

constructor TWhatIsThis.Create;
begin
  inherited;
  FMouseHook := TMouseHook.Create (Self);
  FMouseHook.OnMouseProc := MouseHookMouseProc;
  if not (csDesigning in ComponentState)
    then FMouseHook.InstallHook;
end;

destructor TWhatIsThis.Destroy;
begin
  FMouseHook.Free;
  inherited;
end;

function TWhatIsThis.MouseHookMouseProc;
var
  Control : TControl;
  ShowHelp : Boolean;
  HelpContext : THelpContext;
  function GetSubControl (AControl : TWinControl; p : TPoint) : TControl;
  var
    SubControl : TControl;
    p2 : TPoint;
  begin
    p2 := AControl.ScreenToClient (p);
    SubControl := AControl.ControlAtPos (p2, True, True);
    if SubControl <> nil
      then if SubControl is TWinControl
        then Result := GetSubControl (SubControl as TWinControl, p)
        else Result := SubControl
      else Result := AControl;
  end;
  function GetHelpContext (Control : TControl) : THelpContext;
  var
    PropInfo : PPropInfo;
  begin
    PropInfo := GetPropInfo (Control, 'HelpContext');
    if PropInfo <> nil
      then
      begin
        Result := GetOrdProp (Control, PropInfo);
        if (Result = 0) and (Control.Parent <> nil)
          then Result := GetHelpContext (Control.Parent);
      end
      else if Control.Parent <> nil
        then Result := GetHelpContext (Control.Parent)
        else Result := 0;
  end;
begin
  if FActive
    then
    begin
      if wParam = WM_LBUTTONDOWN
        then
        begin
          Control := FindControl (FMouseHook.MouseHookStruct.hwnd);
          if Control <> nil
            then
            begin
              with FMouseHook.MouseHookStruct.pt do
                Control := GetSubControl (Control as TWinControl, Point (x, y));
              HelpContext := GetHelpContext (Control);
              if HelpContext <> 0
                then
                begin
                  ShowHelp := True;
                  if Assigned (FOnWhatIsThis)
                    then FOnWhatIsThis (Self, Control, HelpContext, ShowHelp);
                  if ShowHelp
                    then Application.HelpContext (HelpContext);
                end
                else if Assigned (FOnNoHelpContext)
                  then FOnNoHelpContext (Self, Control);
            end;
          Result := 1;
          Active := False;
        end
        else Result := 1;
    end
    else Result := 0;
end;

procedure TWhatIsThis.SetActive;
begin
  if Value <> FActive
    then
    begin
      FActive := Value;
      if FActive
        then Screen.Cursor := crHelp
        else Screen.Cursor := crDefault;
      if Assigned (FOnActiveChanged)
        then FOnActiveChanged (Self);
    end;
end;

{ Register }

procedure Register;
begin
  RegisterComponents('Misc2', [TWhatIsThis]);
end;

end.
